# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the custom exceptions for the supervisor. """

import functools
import sys
from botocore.exceptions import ClientError
from faassupervisor.logger import get_logger


def exception():
    """ A decorator that wraps the passed in function and logs exceptions. """

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)

            except ClientError as cerr:
                print(f"There was an exception in {func.__name__}")
                print(cerr.response['Error']['Message'])
                get_logger().error(cerr)
                sys.exit(1)

            except FaasSupervisorError as fse:
                print(fse.args[0])
                get_logger().error(fse)
                if 'Warning' in fse.__class__.__name__:
                    get_logger().warning(fse)
                # Finish the execution if it's an error
                if 'Error' in fse.__class__.__name__:
                    get_logger().error(fse)
                    sys.exit(1)

        return wrapper

    return decorator


class FaasSupervisorError(Exception):
    """
    The base exception class for exceptions.

    :ivar msg: The descriptive message associated with the error.
    """
    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


################################################
# #             GENERAL EXCEPTIONS             ##
################################################
class InvalidPlatformError(FaasSupervisorError):
    """
    The binary is not launched on a Linux platform

    """
    fmt = "This binary only works on a Linux Platform.\nTry executing the Python version."


class InvalidSupervisorTypeError(FaasSupervisorError):
    """
    The supervisor type is not in the allowed list

    """
    fmt = "The supervisor type '{sup_typ}' is not allowed."


class InvalidStoragePathTypeError(FaasSupervisorError):
    """
    The storage path type is not in the allowed list

    """
    fmt = "The storage path type '{storage_type}' is not allowed."


class ContainerImageNotFoundError(FaasSupervisorError):
    """
    The container image is no specified

    """
    fmt = "Container image id is not specified."

class ContainerTimeoutExpiredWarning(FaasSupervisorError):
    """
    The udocker containers has exceeded the defined execution time.

    """
    fmt = "Container timeout expired.\nContainer execution stopped."

class NoLambdaContextError(FaasSupervisorError):
    """
    No context was provided for the lambda instance.

    """
    fmt = "No context found in the Lambda environment."

class UnknowStorageEventWarning(FaasSupervisorError):
    """
    Unknown storage event detected

    """
    fmt = "Unknown storage event detected."

################################################
# #        STORAGE PROVIDER EXCEPTIONS         ##
################################################
class InvalidStorageProviderError(FaasSupervisorError):
    """
    The storage provider type is not valid.

    """
    fmt = "Invalid storage provider type defined: '{storage_type}'."


class NoStorageProviderDefinedWarning(FaasSupervisorError):
    """
    There is no storage provider defined.

    """
    fmt = "There is no storage provider defined for this function execution."


class NoInputStorageProviderDefinedWarning(NoStorageProviderDefinedWarning):
    """
    There is no input storage provider defined.

    """
    fmt = "There is no input storage provider defined for this function execution."


class NoOutputStorageProviderDefinedWarning(NoStorageProviderDefinedWarning):
    """
    There is no output storage provider defined.

    """
    fmt = "There is no output storage provider defined for this function execution."


class StorageTypeError(FaasSupervisorError):
    """
    The storage type defined is not allowed

    """
    fmt = "The storage type '{auth_type}' is not allowed."



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes to parse, store and manage storage authentication information."""
from collections import namedtuple
from faassupervisor.utils import SysUtils


class AuthData():
    """Stores provider authentication values."""

    def __init__(self, storage_id, storage_type):
        self.storage_id = storage_id
        self.type = storage_type
        self.creds = {}

    def set_credential(self, key, val):
        """Store authentication credentials like USER|PASS|TOKEN|SPACE|HOST."""
        self.creds[key] = val

    def get_credential(self, key):
        """Return authentication credentials previously stored."""
        return self.creds.get(key, "")


class StorageAuth():
    """Parses all the provider authentication variables."""

    def __init__(self):
        self.auth_id = {}
        self.auth_type = {}

    def read_storage_providers(self):
        """Reads the global variables to create the providers needed.

        Variable schema:  STORAGE_AUTH_$1_$2_$3
        $1: MINIO | S3 | ONEDATA
        $2: USER | PASS | TOKEN | SPACE | HOST
        $3: STORAGE_ID (Specified in the function definition file,
                        is unique for each storage defined)

        e.g.: STORAGE_AUTH_MINIO_USER_12345
        """
        # Remove the prefix 'STORAGE_AUTH_'
        env_vars = SysUtils.get_filtered_env_vars("STORAGE_AUTH_")
        # type = MINIO | S3 | ONEDATA ...
        # cred = USER | PASS | TOKEN ...
        provider_info = namedtuple('provider_info', ['type', 'cred', 'id'])
        for env_key, env_val in env_vars.items():
            # Don't split past the id
            # MINIO_USER_123_45 -> *[MINIO, USER, 123_45]
            prov_info = provider_info(*env_key.split("_", 2))
            # Link ID with TYPE
            if prov_info.id not in self.auth_id:
                self.auth_id[prov_info.id] = prov_info.type
            if prov_info.type not in self.auth_type:
                # Link TYPE with AUTH data
                self.auth_type[prov_info.type] = AuthData(prov_info.id, prov_info.type)
            self.auth_type[prov_info.type].set_credential(prov_info.cred, env_val)

    def get_auth_data_by_stg_type(self, storage_type):
        """Returns the authentication credentials previously stored."""
        return self.auth_type.get(storage_type)

    def get_data_by_stg_id(self, storage_id):
        """Returns the authentication credentials previously stored."""
        prov_type = self.auth_id.get(storage_id)
        return self.get_auth_data_by_stg_type(prov_type)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the Onedata storage provider. """

import requests
from faassupervisor.logger import get_logger
from faassupervisor.storage.providers import DefaultStorageProvider
from faassupervisor.utils import SysUtils, FileUtils


class Onedata(DefaultStorageProvider):
    """ Class that manages downloads and uploads from Onedata. """

    _CDMI_PATH = 'cdmi'
    _TYPE = 'ONEDATA'

    def __init__(self, stg_auth, stg_path):
        super().__init__(stg_auth, stg_path)
        self._set_onedata_environment()

    def _set_onedata_environment(self):
        self.oneprovider_space = self.stg_auth.get_credential('SPACE')
        self.oneprovider_host = self.stg_auth.get_credential('HOST')
        self.headers = {'X-Auth-Token': self.stg_auth.get_credential('TOKEN')}

    def download_file(self, parsed_event, input_dir_path):
        """ Downloads the file from the space of Onedata and
        returns the path were the download is placed. """
        file_download_path = ""
        url = f"https://{self.oneprovider_host}/{self._CDMI_PATH}{parsed_event.object_key}"
        get_logger().info("Downloading item from host '%s' with key '%s'",
                          self.oneprovider_host,
                          parsed_event.object_key)
        response = requests.get(url, headers=self.headers)
        if response.status_code == 200:
            file_download_path = SysUtils.join_paths(input_dir_path, parsed_event.file_name)
            FileUtils.create_file_with_content(file_download_path, response.content, mode='wb')

            get_logger().info("Successful download of file '%s' with key '%s' in path '%s'",
                              parsed_event.file_name,
                              parsed_event.object_key,
                              file_download_path)
        else:
            get_logger().error("File '%s' download from Onedata host '%s' failed!",
                               parsed_event.file_name,
                               self.oneprovider_host)
        return file_download_path

    def upload_file(self, file_path, file_name):
        url = (f'https://{self.oneprovider_host}/{self._CDMI_PATH}/'
               f'{self.oneprovider_space}/{self.stg_path}/{file_name}')
        get_logger().info("Uploading file '%s' to '%s/%s'",
                          file_name,
                          self.oneprovider_space,
                          self.stg_path)
        with open(file_path, 'rb') as data:
            response = requests.put(url, data=data, headers=self.headers)
            if response.status_code not in [201, 202, 204]:
                get_logger().error("Upload failed. Status code: %s", response.status_code)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the Minio storage provider. """

import boto3
from faassupervisor.logger import get_logger
from faassupervisor.storage.providers import DefaultStorageProvider
from faassupervisor.utils import SysUtils


class Minio(DefaultStorageProvider):
    """ Class that manages downloads and uploads from Minio. """

    _DEFAULT_MINIO_ENDPOINT = 'http://minio-service.minio:9000'
    _TYPE = 'MINIO'

    def _get_client(self):
        """Return Minio client with user configuration."""
        endpoint = SysUtils.get_env_var('MINIO_ENDPOINT')
        if not endpoint:
            endpoint = self._DEFAULT_MINIO_ENDPOINT
        return boto3.client('s3', endpoint_url=endpoint,
                            aws_access_key_id=self.stg_auth.get_credential('USER'),
                            aws_secret_access_key=self.stg_auth.get_credential('PASS'))

    def download_file(self, parsed_event, input_dir_path):
        """Downloads a file from a minio bucket."""
        file_download_path = SysUtils.join_paths(input_dir_path, parsed_event.file_name)
        get_logger().info("Downloading item from bucket '%s' with key '%s'",
                          parsed_event.bucket_name,
                          parsed_event.file_name)

        with open(file_download_path, 'wb') as data:
            self._get_client().download_fileobj(parsed_event.bucket_name,
                                                parsed_event.file_name,
                                                data)
        get_logger().info("Successful download of file '%s' from bucket '%s' in path '%s'",
                          parsed_event.file_name,
                          parsed_event.bucket_name,
                          file_download_path)
        return file_download_path

    def upload_file(self, file_path, file_name):
        """Uploads a file to a minio bucket."""
        get_logger().info("Uploading file '%s' to bucket '%s'", file_name, self.stg_path)
        with open(file_path, 'rb') as data:
            self._get_client().upload_fileobj(data, self.stg_path, file_name)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the S3 storage provider. """

import boto3
from faassupervisor.logger import get_logger
from faassupervisor.storage.providers import DefaultStorageProvider
from faassupervisor.utils import SysUtils


def _get_client():
    """Returns S3 client with default configuration."""
    return boto3.client('s3')


def _set_file_acl(bucket_name, file_key):
    obj = boto3.resource('s3').Object(bucket_name, file_key)
    obj.Acl().put(ACL='public-read')


class S3(DefaultStorageProvider):
    """Class that manages downloads and uploads from S3."""

    _TYPE = 'S3'

    def _get_file_key(self, file_name):
        stg_path = self.stg_path.split('/', 1)
        # Path format => stg_path: bucket/<folder-path>
        # Last part is optional
        if len(stg_path) > 1:
            # There is a folder defined
            # Set the folder in the file path
            file_key = f"{stg_path[1]}/{file_name}"
        else:
            # Set the default file path
            file_key = (f"{SysUtils.get_env_var('AWS_LAMBDA_FUNCTION_NAME')}/output/"
                        f"{SysUtils.get_env_var('AWS_LAMBDA_REQUEST_ID')}/{file_name}")
        return file_key

    def _get_bucket_name(self):
        return self.stg_path.split("/")[0]

    def download_file(self, parsed_event, input_dir_path):
        """ Downloads the file from the S3 bucket and
        returns the path were the download is placed. """
        file_download_path = SysUtils.join_paths(input_dir_path, parsed_event.file_name)
        get_logger().info("Downloading item from bucket '%s' with key '%s'",
                          parsed_event.bucket_name,
                          parsed_event.object_key)
        with open(file_download_path, 'wb') as data:
            _get_client().download_fileobj(parsed_event.bucket_name,
                                           parsed_event.object_key,
                                           data)
        get_logger().info("Successful download of file '%s' from bucket '%s' in path '%s'",
                          parsed_event.object_key,
                          parsed_event.bucket_name,
                          file_download_path)
        return file_download_path

    def upload_file(self, file_path, file_name):
        file_key = self._get_file_key(file_name)
        bucket_name = self._get_bucket_name()
        get_logger().info("Uploading file '%s' to bucket '%s'", file_key, bucket_name)
        with open(file_path, 'rb') as data:
            _get_client().upload_fileobj(data, bucket_name, file_key)

        get_logger().info("Changing ACLs for public-read for object in bucket '%s' with key '%s'",
                          bucket_name,
                          file_key)
        _set_file_acl(bucket_name, file_key)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.storage module and classes."""

import unittest
from unittest import mock
from unittest.mock import call
from collections import namedtuple
import faassupervisor.storage as storage
from faassupervisor.exceptions import InvalidStorageProviderError
from faassupervisor.storage.auth import AuthData, StorageAuth
from faassupervisor.storage.providers.local import Local
from faassupervisor.storage.providers.minio import Minio
from faassupervisor.events.minio import MinioEvent
from faassupervisor.storage.providers.onedata import Onedata
from faassupervisor.storage.providers.s3 import S3, _get_client
from faassupervisor.events.apigateway import ApiGatewayEvent
from faassupervisor.events.s3 import S3Event
from faassupervisor.events.onedata import OnedataEvent

# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=protected-access


class StorageModuleTest(unittest.TestCase):

    auth = namedtuple("auth", ["type"])

    def test_create_provider_local(self):
        prov = storage.create_provider(())
        self.assertEqual(prov.get_type(), "LOCAL")

    def test_create_provider_minio(self):
        prov = storage.create_provider(self.auth("MINIO"))
        self.assertEqual(prov.get_type(), "MINIO")

    @mock.patch("faassupervisor.storage.providers.onedata.Onedata._set_onedata_environment")
    def test_create_provider_onedata(self, onepatch):
        prov = storage.create_provider(self.auth("ONEDATA"))
        onepatch.assert_called_once()
        self.assertEqual(prov.get_type(), "ONEDATA")

    def test_create_provider_s3(self):
        prov = storage.create_provider(self.auth("S3"))
        self.assertEqual(prov.get_type(), "S3")

    def test_create_provider_invalid(self):
        with self.assertRaises(InvalidStorageProviderError):
            storage.create_provider(self.auth("ERROR"))

    @mock.patch("faassupervisor.storage.providers.s3.S3.download_file")
    def test_download_input(self, mock_s3):
        prov = storage.create_provider(self.auth("S3"))
        storage.download_input(prov, {}, '/tmp/test')
        mock_s3.assert_called_once_with({}, '/tmp/test')

    @mock.patch("faassupervisor.utils.FileUtils.get_all_files_in_dir")
    @mock.patch("faassupervisor.storage.providers.s3.S3.upload_file")
    def test_upload_output(self, mock_s3, mock_utils):
        mock_utils.return_value = ['/tmp/test/f1', '/tmp/test/k1/f2']
        prov = storage.create_provider(self.auth("S3"))
        storage.upload_output(prov, '/tmp/test')
        mock_s3.call_count = 2
        mock_s3.mock_call()[0] = call('/tmp/test/f1', 'f1')
        mock_s3.mock_call()[1] = call('/tmp/test/k1/f2', 'k1/f2')

    def test_get_output_paths(self):
        # Mock environment variables
        with mock.patch.dict('os.environ',
                             {"STORAGE_PATH_OUTPUT_1" : "tmp1", "STORAGE_PATH_OUTPUT_2" : "tmp1"},
                             clear=True):
            result = storage.get_output_paths()
            storage_path = namedtuple('storage_path', ['id', 'path'])
            self.assertEqual(result,
                             [storage_path(id='1', path='tmp1'),
                              storage_path(id='2', path='tmp1')])


class AuthDataTest(unittest.TestCase):

    def test_create_auth_data(self):
        auth = AuthData('1', 'LOCAL')
        self.assertEqual(auth.storage_id, '1')
        self.assertEqual(auth.type, 'LOCAL')
        self.assertEqual(auth.creds, {})

    def test_set_auth_data_credential(self):
        auth = AuthData('1', 'LOCAL')
        auth.set_credential('K1', 'V1')
        self.assertEqual(auth.creds, {'K1':'V1'})

    def test_get_auth_data_credential(self):
        auth = AuthData('1', 'LOCAL')
        auth.set_credential('K1', 'V1')
        self.assertEqual(auth.get_credential('K1'), 'V1')
        self.assertEqual(auth.get_credential('K11'), '')


class StorageAuthTest(unittest.TestCase):

    def test_create_storage_auth(self):
        stga = StorageAuth()
        self.assertEqual(stga.auth_id, {})
        self.assertEqual(stga.auth_type, {})

    @mock.patch('faassupervisor.storage.auth.AuthData')
    def test_read_storage_providers(self, mock_auth):
        # Mock environment variables
        with mock.patch.dict('os.environ',
                             {"STORAGE_AUTH_S3_USER_1" : "u1", "STORAGE_AUTH_S3_PASS_1" : "p1"},
                             clear=True):
            stga = StorageAuth()
            stga.read_storage_providers()
            self.assertEqual(stga.auth_id, {'1' : 'S3'})
            mock_auth.assert_called_once()
            self.assertEqual(mock_auth.mock_calls[0], call('1', 'S3'))
            self.assertEqual(mock_auth.mock_calls[1], call().set_credential('USER', 'u1'))
            self.assertEqual(mock_auth.mock_calls[2], call().set_credential('PASS', 'p1'))

    def test_get_auth_data_by_stg_type(self):
        # Mock environment variables
        with mock.patch.dict('os.environ',
                             {"STORAGE_AUTH_S3_USER_1" : "u1", "STORAGE_AUTH_S3_PASS_1" : "p1"},
                             clear=True):
            stga = StorageAuth()
            stga.read_storage_providers()
            self.assertEqual(stga.get_auth_data_by_stg_type('S3').get_credential('USER'), 'u1')
            self.assertEqual(stga.get_auth_data_by_stg_type('S3').get_credential('PASS'), 'p1')

    def test_get_data_by_stg_id(self):
        # Mock environment variables
        with mock.patch.dict('os.environ',
                             {"STORAGE_AUTH_S3_USER_1" : "u1", "STORAGE_AUTH_S3_PASS_1" : "p1"},
                             clear=True):
            stga = StorageAuth()
            stga.read_storage_providers()
            self.assertEqual(stga.get_data_by_stg_id('1').get_credential('USER'), 'u1')
            self.assertEqual(stga.get_data_by_stg_id('1').get_credential('PASS'), 'p1')


class LocalStorageTest(unittest.TestCase):

    def test_create_local_storage(self):
        stg = Local('auth', 'path')
        self.assertEqual(stg.stg_auth, 'auth')
        self.assertEqual(stg.stg_path, 'path')
        self.assertEqual(stg.get_type(), 'LOCAL')

    def test_download_file(self):
        stg = Local('auth', 'path')
        parsed_event = mock.Mock(spec=ApiGatewayEvent)
        stg.download_file(parsed_event, '/tmp/local')
        parsed_event.save_event.assert_called_once_with('/tmp/local')


class MinioStorageTest(unittest.TestCase):

    def _get_minio_class_and_auth(self):
        auth = mock.Mock(spec=AuthData)
        auth.get_credential.return_value = 'test'
        return (Minio(auth, 'minio_path'), auth)

    def test_create_minio_storage(self):
        stg, auth = self._get_minio_class_and_auth()
        self.assertEqual(stg.stg_auth, auth)
        self.assertEqual(stg.stg_path, 'minio_path')
        self.assertEqual(stg.get_type(), 'MINIO')

    @mock.patch('boto3.client')
    def test_get_client_default_endpoint(self, mock_boto):
        stg, auth = self._get_minio_class_and_auth()
        stg._get_client()
        self.assertEqual(auth.get_credential.mock_calls[0], call('USER'))
        self.assertEqual(auth.get_credential.mock_calls[1], call('PASS'))
        mock_boto.assert_called_once_with('s3',
                                          aws_access_key_id='test',
                                          aws_secret_access_key='test',
                                          endpoint_url='http://minio-service.minio:9000')

    @mock.patch('boto3.client')
    def test_get_client(self, mock_boto):
        # Mock environment variables
        with mock.patch.dict('os.environ', {"MINIO_ENDPOINT" : "test_endpoint"}, clear=True):
            stg, auth = self._get_minio_class_and_auth()
            stg._get_client()
            self.assertEqual(auth.get_credential.mock_calls[0], call('USER'))
            self.assertEqual(auth.get_credential.mock_calls[1], call('PASS'))
            mock_boto.assert_called_once_with('s3',
                                              aws_access_key_id='test',
                                              aws_secret_access_key='test',
                                              endpoint_url='test_endpoint')

    @mock.patch('boto3.client')
    def test_download_file(self, mock_boto):
        stg, _ = self._get_minio_class_and_auth()
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            # Create mock event and event properties
            event = mock.Mock(spec=MinioEvent)
            type(event).bucket_name = mock.PropertyMock(return_value='minio_bucket')
            type(event).file_name = mock.PropertyMock(return_value='minio_file')
            file_path = stg.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/minio_file')
            # Check file writing call
            mopen.assert_called_once_with('/tmp/input/minio_file', 'wb')
            # Check boto client download call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().download_fileobj('minio_bucket',
                                                     'minio_file',
                                                     mopen.return_value))

    @mock.patch('boto3.client')
    def test_upload_file(self, mock_boto):
        stg, _ = self._get_minio_class_and_auth()
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            stg.upload_file('/tmp/output/processed.jpg', 'processed.jpg')
            # Check file reading call
            mopen.assert_called_once_with('/tmp/output/processed.jpg', 'rb')
            # Check boto client upload call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().upload_fileobj(mopen.return_value,
                                                   'minio_path',
                                                   'processed.jpg'))


class S3StorageTest(unittest.TestCase):

    def _get_s3_class_and_auth(self, path=None):
        auth = mock.Mock(spec=AuthData)
        auth.get_credential.return_value = 'test'
        s3_path = 's3_path'
        if path:
            s3_path = path
        return (S3(auth, s3_path), auth)

    def test_create_s3_storage(self):
        stg, auth = self._get_s3_class_and_auth()
        self.assertEqual(stg.stg_auth, auth)
        self.assertEqual(stg.stg_path, 's3_path')
        self.assertEqual(stg.get_type(), 'S3')

    def test_get_file_key_wo_folder(self):
        stg, _ = self._get_s3_class_and_auth()
        # Mock environment variables
        with mock.patch.dict('os.environ',
                             {"AWS_LAMBDA_FUNCTION_NAME" : "func_name",
                              "AWS_LAMBDA_REQUEST_ID" : "req_id"},
                             clear=True):
            res = stg._get_file_key('test.jpg')
            self.assertEqual(res, "func_name/output/req_id/test.jpg")

    def test_get_file_key_w_folder(self):
        stg, _ = self._get_s3_class_and_auth(path='bucket/f1/f2')
        res = stg._get_file_key('test.jpg')
        self.assertEqual(res, "f1/f2/test.jpg")

    def test_get_bucket_name_wo_folder(self):
        stg, _ = self._get_s3_class_and_auth()
        res = stg._get_bucket_name()
        self.assertEqual(res, 's3_path')

    def test_get_bucket_name_w_folder(self):
        stg, _ = self._get_s3_class_and_auth(path='bucket/f1/f2')
        res = stg._get_bucket_name()
        self.assertEqual(res, 'bucket')

    @mock.patch('boto3.client')
    def test_get_client(self, mock_boto):
        _get_client()
        mock_boto.assert_called_once_with('s3')

    @mock.patch('boto3.client')
    def test_download_file(self, mock_boto):
        stg, _ = self._get_s3_class_and_auth()
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            # Mock S3 event and event properties
            event = mock.Mock(spec=S3Event)
            type(event).bucket_name = mock.PropertyMock(return_value='s3_bucket')
            type(event).file_name = mock.PropertyMock(return_value='s3_file')
            type(event).object_key = mock.PropertyMock(return_value='s3_file_key')
            file_path = stg.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/s3_file')
            # Check file writing call
            mopen.assert_called_once_with('/tmp/input/s3_file', 'wb')
            # Check boto client download call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().download_fileobj('s3_bucket',
                                                     's3_file_key',
                                                     mopen.return_value))

    @mock.patch('boto3.resource')
    @mock.patch('boto3.client')
    def test_upload_file(self, mock_boto_client, mock_boto_resource):
        stg, _ = self._get_s3_class_and_auth()
        # Mock environment variables
        with mock.patch.dict('os.environ',
                             {"AWS_LAMBDA_FUNCTION_NAME" : "func_name",
                              "AWS_LAMBDA_REQUEST_ID" : "req_id"},
                             clear=True):
            # Mock file management
            mopen = mock.mock_open()
            with mock.patch('builtins.open', mopen, create=True):
                stg.upload_file('/tmp/output/processed.jpg', 'processed.jpg')
                # Check file read
                mopen.assert_called_once_with('/tmp/output/processed.jpg', 'rb')
                # Check boto client upload call
                self.assertEqual(mock_boto_client.mock_calls[1],
                                 call().upload_fileobj(mopen.return_value,
                                                       's3_path',
                                                       'func_name/output/req_id/processed.jpg'))
                # Check resource creation
                mock_boto_resource.assert_called_once_with('s3')
                # Check resource call
                self.assertEqual(mock_boto_resource.mock_calls[1],
                                 call().Object('s3_path', 'func_name/output/req_id/processed.jpg'))
                # Check setting resource properties
                self.assertEqual(mock_boto_resource.mock_calls[3],
                                 call().Object().Acl().put(ACL='public-read'))


class OnedataStorageTest(unittest.TestCase):

    def _get_onedata_class_and_auth(self, path=None):
        auth = mock.Mock(spec=AuthData)
        auth.get_credential.return_value = 'test'
        one_path = 'onedata_path'
        if path:
            one_path = path
        return (Onedata(auth, one_path), auth)

    def test_create_onedata_storage(self):
        stg, auth = self._get_onedata_class_and_auth()
        self.assertEqual(stg.stg_auth, auth)
        self.assertEqual(stg.stg_auth.mock_calls[0], call.get_credential('SPACE'))
        self.assertEqual(stg.stg_auth.mock_calls[1], call.get_credential('HOST'))
        self.assertEqual(stg.stg_auth.mock_calls[2], call.get_credential('TOKEN'))
        self.assertEqual(stg.stg_path, 'onedata_path')
        self.assertEqual(stg.get_type(), 'ONEDATA')

    @mock.patch('requests.get')
    def test_download_file(self, mock_requests):
        stg, _ = self._get_onedata_class_and_auth()
        # Mock requests.get response
        response = namedtuple('response', ['content', 'status_code'])
        mock_requests.return_value = response('test response', 200)
        # Create mock event
        event = mock.Mock(spec=OnedataEvent)
        type(event).file_name = mock.PropertyMock(return_value='onedata_file')
        type(event).object_key = mock.PropertyMock(return_value='/onedata_file_key')
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            file_path = stg.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/onedata_file')
            # Check request to onedata endpoint
            mock_requests.assert_called_once_with('https://test/cdmi/onedata_file_key',
                                                  headers={'X-Auth-Token': 'test'})
            # Check file writing
            mopen.assert_called_once_with('/tmp/input/onedata_file', 'wb')

    @mock.patch('requests.put')
    def test_upload_file(self, mock_requests):
        stg, _ = self._get_onedata_class_and_auth()
        # Mock requests.put response
        response = namedtuple('response', ['status_code'])
        mock_requests.return_value = response(202)
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            stg.upload_file('/tmp/output/onedata_file', 'onedata_file')
            # Check request to onedata endpoint
            mock_requests.assert_called_once_with(
                'https://test/cdmi/test/onedata_path/onedata_file',
                data=mopen.return_value,
                headers={'X-Auth-Token': 'test'})
            # Check file writing
            mopen.assert_called_once_with('/tmp/output/onedata_file', 'rb')



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines a generic storage
provider to be used by the generic supervisor."""

from collections import namedtuple
from faassupervisor.utils import SysUtils
from faassupervisor.storage.providers.minio import Minio
from faassupervisor.storage.providers.onedata import Onedata
from faassupervisor.storage.providers.local import Local
from faassupervisor.storage.providers.s3 import S3
from faassupervisor.logger import get_logger
from faassupervisor.utils import FileUtils
from faassupervisor.exceptions import InvalidStorageProviderError


def create_provider(storage_auth, storage_path=None):
    """Returns the storage provider needed
    based on the authentication type defined.

    If not storage auth provided, use local storage."""
    if not storage_auth:
        provider = Local(storage_auth, storage_path)
    elif storage_auth.type == 'MINIO':
        provider = Minio(storage_auth, storage_path)
    elif storage_auth.type == 'ONEDATA':
        provider = Onedata(storage_auth, storage_path)
    elif storage_auth.type == 'S3':
        provider = S3(storage_auth, storage_path)
    else:
        raise InvalidStorageProviderError(storage_type=storage_auth.type)
    return provider


def download_input(storage_provider, parsed_event, input_dir_path):
    """Receives the event where the file information is and
    the tmp_dir_path where to store the downloaded file.

    Returns the file path where the file is downloaded."""
    return storage_provider.download_file(parsed_event, input_dir_path)


def upload_output(storage_provider, output_dir_path):
    """Receives the tmp_dir_path where the files to upload are stored and
    uploads all the files found there."""

    get_logger().info("Searching for files to upload in folder '%s'", output_dir_path)
    output_files = FileUtils.get_all_files_in_dir(output_dir_path)
    get_logger().info("Found the following files to upload: '%s'", output_files)
    for file_path in output_files:
        file_name = file_path.replace(f"{output_dir_path}/", "")
        storage_provider.upload_file(file_path, file_name)


def get_output_paths():
    """Returns the defined output providers.

    Reads the global variables to create the providers needed.
    Variable schema: STORAGE_PATH_$1_$2
                     $1: INPUT | OUTPUT
                     $2: STORAGE_ID (Specified in the function definition file,
                                     is unique for each storage defined)
    e.g.: STORAGE_PATH_INPUT_12345
    """
    get_logger().info("Reading output path variables")
    env_vars = SysUtils.get_filtered_env_vars("STORAGE_PATH_")
    storage_path = namedtuple('storage_path', ['id', 'path'])
    # Remove the storage type 'OUTPUT_' and store only the id and the path
    # Store a tuple, so the information can't be modified
    return [storage_path(env_key[7:], env_val) for env_key, env_val in env_vars.items()
            if env_key.startswith('OUTPUT_')]
